from typing import Any, Callable, TypeVar, ParamSpec
from os import system
from sys import platform


P = ParamSpec("P")
R = TypeVar("R")

FUNCTION = Callable[..., object]


class FunctionRegistry:
    def __init__(self, *valid_keys: str):
        self._events: dict[str, FUNCTION] = {}

        self._validKeys = valid_keys

    def __call__(self, arg: str | FUNCTION) -> FUNCTION | Callable[[FUNCTION], FUNCTION]:
        if isinstance(arg, str):
            def decorator(func: FUNCTION) -> FUNCTION:
                return self._register(arg, func)
            return decorator

        else:
            return self._register(arg.__name__, arg)

    def __getattr__(self, name: str) -> FUNCTION:
        if name in self.__dict__.keys() or name in self.__dir__():
            return self.__dict__[name]

        elif name in self._events:
            return self._events[name]

        else:
            return self._DefaultFunction()

    class _DefaultFunction:
        def __call__(self, *args: Any, **kwds: Any):
            return None

        def __bool__(self) -> bool:
            return False

    def _register(self, name: str, func: FUNCTION) -> FUNCTION:
        if name in self.__dict__.keys() or name in self.__dir__():
            raise ValueError(
                f"The name {name} is used by the class {type(self).__name__}")

        elif self._validKeys and name not in self._validKeys:
            raise ValueError(f"{name} is not a valid key")

        self._events[name] = func
        return func


def clear_console():
    if platform.startswith("linux") or platform == "darwin":
        system("clear")

    elif platform.startswith("win"):
        system("cls")


__all__ = ["FunctionRegistry", "clear_console", "FUNCTION"]


if __name__ == '__main__':
    clear_console()

    class Mailbox:
        def __init__(self):
            self.event = FunctionRegistry("on_mail", "on_spam")

    mailbox = Mailbox()

    @mailbox.event
    def on_mail(message: str):
        print(message)

    on_mail_event = mailbox.event.on_mail
    if on_mail_event:
        on_mail_event("Hello, World!")

    @mailbox.event("on_spam")
    def spam_handler():
        print("Spam")

    mailbox.event.on_spam()



from tkinter import Tk, Frame, Misc, Canvas, Scrollbar, Event, Misc
from typing import Callable, Optional, Any
from sys import platform


class APMMisc(Misc):
    def set_rows(self, *row_weights: int):
        if len(row_weights) == 1:
            row_weights = (1, ) * row_weights[0]

        _, r = self.grid_size()
        m = max(r, len(row_weights))

        for i in range(m):
            if i < r and i >= len(row_weights):
                self.grid_rowconfigure(i, weight=0)

            elif i < len(row_weights):
                self.grid_rowconfigure(i, weight=row_weights[i])

    def set_columns(self, *column_weights: int):
        if len(column_weights) == 1:
            column_weights = (1, ) * column_weights[0]

        c, _ = self.grid_size()
        m = max(c, len(column_weights))

        for i in range(m):
            if i < c and i >= len(column_weights):
                self.grid_columnconfigure(i, weight=0)

            elif i < len(column_weights):
                self.grid_columnconfigure(i, weight=column_weights[i])

    def set_rows_columns(self, row_weights: tuple[int, ...] | int, column_weights: tuple[int, ...] | int):
        if isinstance(row_weights, int):
            row_weights = (1, ) * row_weights

        if isinstance(column_weights, int):
            column_weights = (1, ) * column_weights

        self.set_rows(*row_weights)
        self.set_columns(*column_weights)

    def rows(self) -> int:
        return self.grid_size()[1]

    def columns(self) -> int:
        return self.grid_size()[0]


class APMTk(Tk, APMMisc):
    def __init__(self, title: Optional[str] = None, resize: tuple[int | float, int | float] | None = (1/2, 1/2), center: bool = True, *args, **kwargs):
        # AI generated
        if platform == "win32":
            try:
                from ctypes import windll  # type: ignore
                windll.shcore.SetProcessDpiAwareness(1)
            except Exception:
                try:
                    windll.user32.SetProcessDPIAware()
                except Exception:
                    pass
        #

        super().__init__(*args, **kwargs)

        if title is not None:
            self.title(title)

        if center:
            self.center()

        if resize is not None:
            self.resize(*resize)

        self.rows: Callable[[], int] = lambda: self.grid_size()[1]
        self.columns: Callable[[], int] = lambda: self.grid_size()[0]

    def resize(self, newWidth: int | float, newHeight: int | float):
        # self.update_idletasks()

        screen_width = self.winfo_screenwidth()
        screen_height = self.winfo_screenheight()

        if isinstance(newWidth, float):
            newWidth = int(screen_width * newWidth)

        if isinstance(newHeight, float):
            newHeight = int(screen_height * newHeight)

        newX = (screen_width - newWidth) // 2
        newY = (screen_height - newHeight) // 2

        self.geometry(f"{newWidth}x{newHeight}+{newX}+{newY}")

    def center(self):
        # self.update_idletasks()

        screen_width = self.winfo_screenwidth()
        screen_height = self.winfo_screenheight()

        width = self.winfo_width()
        height = self.winfo_height()

        x = (screen_width - width) // 2
        y = (screen_height - height) // 2

        self.geometry(f"+{x}+{y}")


class APMFrame(Frame, APMMisc):
    ...


class ScrollingFrame(APMFrame):
    def __init__(self, master: Misc | None = None, cnf: dict[str, Any] | None = None, **kwargs):
        self.frame = APMFrame(master)
        self.frame.grid_propagate(False)
        self.frame.set_rows_columns((1,), (1,))

        self.canvas = Canvas(self.frame)
        self.canvas.grid(row=0, column=0, sticky="nsew")

        self.scrollbar = Scrollbar(self.frame, command=self.canvas.yview)
        self.scrollbar.grid(row=0, column=1, sticky="ns")
        self.canvas.configure(yscrollcommand=self.scrollbar.set)

        super().__init__(self.canvas, **kwargs)
        self.id = self.canvas.create_window((0, 0), window=self, anchor="nw")

        self.canvas.bind("<Configure>", self.on_canvas_configure)
        self.bind("<Configure>", self.on_frame_configure)

        self.canvas.bind("<Enter>", lambda _: self._bind_mousewheel())
        self.canvas.bind("<Leave>", lambda _: self._unbind_mousewheel())

    def on_canvas_configure(self, event: Event):
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))
        self.canvas.itemconfig(self.id, width=event.width)
        self.canvas.configure(background=self.cget("background"))

    def on_frame_configure(self, event: Event):
        self.on_canvas_configure(event)
        self.canvas.bind_all("<MouseWheel>", self.mouse_scroll)

    def _bind_mousewheel(self):
        if platform.startswith("win"):
            self.canvas.bind_all("<MouseWheel>", self.mouse_scroll)
        elif platform == "darwin":
            self.canvas.bind_all("<MouseWheel>", self.mouse_scroll)
        else:
            self.canvas.bind_all("<Button-4>", self.mouse_scroll)
            self.canvas.bind_all("<Button-5>", self.mouse_scroll)

    def _unbind_mousewheel(self):
        if platform.startswith("win") or platform == "darwin":
            self.canvas.unbind_all("<MouseWheel>")
        else:
            self.canvas.unbind_all("<Button-4>")
            self.canvas.unbind_all("<Button-5>")

    def mouse_scroll(self, event: Event):
        from tkinter import TclError
        try:
            if platform.startswith("win") or platform == "darwin":
                move = -1 if event.delta > 0 else 1
            else:
                move = -1 if event.num == 4 else 1

            self.canvas.yview_scroll(move, "units")

        except TclError:
            self._unbind_mousewheel()

    def grid(self, **kwargs): self.frame.grid(**kwargs)
    def pack(self, **kwargs): self.frame.pack(**kwargs)


__all__ = ["APMTk", "ScrollingFrame", "APMFrame"]

if __name__ == '__main__':
    from tkinter import Label

    root = APMTk("Test Window")

    root.set_rows(1)
    root.set_columns(1)

    scrolling_frame = ScrollingFrame()
    scrolling_frame.frame.grid(row=0, column=0, sticky="nsew")

    for i in range(1, 100):
        Label(scrolling_frame, text=f"Label {i}").pack()

    root.mainloop()



from socket import socket, AF_INET, SOCK_STREAM, gethostbyname, gethostname, SOL_SOCKET, SO_REUSEADDR
from threading import Thread
from time import sleep
from typing import Any, Callable
from json import dumps, loads
from datetime import datetime
from apm import FunctionRegistry, FUNCTION
from typing import TypedDict, Literal
from uuid import getnode


class Server:
    class ServerFunctionRegistry(FunctionRegistry):
        EVENT = Literal["on_message", "on_client_connect"]

        def __call__(self, arg: EVENT | FUNCTION) -> FUNCTION | Callable[[FUNCTION], FUNCTION]:
            return super().__call__(arg)

    def __init__(self) -> None:
        self.running: bool = False

        self.socket: socket | None = None

        self.event = self.ServerFunctionRegistry()

    def start(self, port: int, backlog: int = 1):
        self.socket = socket(AF_INET, SOCK_STREAM)
        self.socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)

        self.socket.bind(("0.0.0.0", port))
        self.socket.listen(backlog)

        self.running = True

        running_thread = Thread(target=self.keep_alive)
        running_thread.start()

        accept_thread = Thread(target=self.accept_clients, daemon=True)
        accept_thread.start()

    def keep_alive(self):
        while self.running:
            sleep(1)

    def stop(self):
        if self.socket:
            self.running = False
            self.socket.close()

    def accept_clients(self):
        while self.running:
            if self.socket:
                client_socket, _ = self.socket.accept()
                Thread(target=self.handle_client,
                       args=(client_socket, )).start()

    def handle_client(self, client_socket: socket):
        with client_socket:
            self.event.on_client_connect(client_socket)

            while self.running:
                raw_data = client_socket.recv(1024)

                if not raw_data:
                    break

                messages = decode_message(raw_data)

                for message in messages:
                    if not message["message"].strip():
                        break

                    self.event.on_message(client_socket, message)

    def __enter__(self) -> "Server":
        return self

    def __exit__(self, *_):
        self.stop()

    def __del__(self):
        self.stop()


def get_ip() -> str:
    return gethostbyname(gethostname())


class DecodedMessage(TypedDict):
    message: str
    encodeTime: datetime
    decodeTime: datetime
    macId: str


def decode_message(raw_data: bytes) -> list[DecodedMessage]:
    messages = raw_data.decode("utf-8").split("\x00")
    decoded_messages = []

    for m in messages:
        if not m:
            continue

        data = loads(m)

        data["encodeTime"] = datetime.fromisoformat(data["encodeTime"])
        data["decodeTime"] = datetime.now()

        decoded_messages.append(data)

    return decoded_messages


def encode_message(message: Any) -> bytes:
    data = {
        "message": message,
        "encodeTime": datetime.now().isoformat(),
        "macId": f"{getnode():012x}"
    }
    return f"{dumps(data)}\x00".encode("utf-8")


class Client:
    class ClientFunctionRegistry(FunctionRegistry):
        EVENT = Literal["on_message"]

        def __call__(self, arg: EVENT | FUNCTION) -> FUNCTION | Callable[[FUNCTION], FUNCTION]:
            return super().__call__(arg)
        
    def __init__(self):
        self.connected: bool = False

        self.socket: socket | None = None

        self.event = self.ClientFunctionRegistry()

    def connect(self, ip: str, port: int):
        self.socket = socket(AF_INET, SOCK_STREAM)
        self.socket.connect((ip, port))

        self.connected = True

        running_thread = Thread(target=self.keep_alive)
        running_thread.start()

        receive_thread = Thread(target=self.receive_messages, daemon=True)
        receive_thread.start()

    def disconnect(self):
        if self.socket:
            self.connected = False
            self.socket.close()

    def send(self, message: str):
        if self.socket:
            data = encode_message(message)
            self.socket.send(data)

    def __del__(self):
        self.disconnect()

    def __enter__(self) -> "Client":
        return self

    def __exit__(self, *_):
        self.disconnect()

    def receive_messages(self):
        if self.socket:
            while self.connected:
                raw_data = self.socket.recv(1024)

                if not raw_data:
                    break

                messages = decode_message(raw_data)

                for message in messages:
                    self.event.on_message(message)

    def keep_alive(self):
        while self.connected:
            sleep(1)


__all__ = ["Server", "decode_message", "encode_message", "Client"]

if __name__ == '__main__':
    from apm import clear_console

    clear_console()

    server = Server()

    @server.event
    def on_message(client_socket: socket, data: DecodedMessage):
        client_ip, client_port = client_socket.getpeername()
        address = f"{client_ip}:{client_port}"

        send_time = data["encodeTime"]

        print(f"[{send_time}] {address} | {data["message"]}")

    server.start(2542)
